from __future__ import annotations

import warnings
from collections.abc import Iterator
from functools import wraps
from numbers import Number

import numpy as np
from tlz import merge

from dask.array.core import Array
from dask.array.dispatch import empty_lookup, percentile_lookup
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph
from dask.utils import derived_from


@wraps(np.percentile)
def _percentile(a, q, method="linear"):
    n = len(a)
    if not len(a):
        return None, n
    if isinstance(q, Iterator):
        q = list(q)
    if a.dtype.name == "category":
        result = np.percentile(a.cat.codes, q, method=method)
        import pandas as pd

        return pd.Categorical.from_codes(result, a.dtype.categories, a.dtype.ordered), n
    if type(a.dtype).__name__ == "DatetimeTZDtype":
        import pandas as pd

        if isinstance(a, (pd.Series, pd.Index)):
            a = a.values

    if np.issubdtype(a.dtype, np.datetime64):
        values = a
        if type(a).__name__ in ("Series", "Index"):
            a2 = values.astype("i8")
        else:
            a2 = values.view("i8")
        result = np.percentile(a2, q, method=method).astype(values.dtype)
        if q[0] == 0:
            # https://github.com/dask/dask/issues/6864
            result[0] = min(result[0], values.min())
        return result, n
    if not np.issubdtype(a.dtype, np.number):
        method = "nearest"
    return np.percentile(a, q, method=method), n


def _tdigest_chunk(a):
    from crick import TDigest

    t = TDigest()
    t.update(a)

    return t


def _percentiles_from_tdigest(qs, digests):
    from crick import TDigest

    t = TDigest()
    t.merge(*digests)

    return np.array(t.quantile(qs / 100.0))


def percentile(a, q, method="linear", internal_method="default", **kwargs):
    """Approximate percentile of 1-D array

    Parameters
    ----------
    a : Array
    q : array_like of float
        Percentile or sequence of percentiles to compute, which must be between
        0 and 100 inclusive.
    method : {'linear', 'lower', 'higher', 'midpoint', 'nearest'}, optional
        The interpolation method to use when the desired percentile lies
        between two data points ``i < j``.

        - 'linear': ``i + (j - i) * fraction``, where ``fraction``
          is the fractional part of the index surrounded by ``i``
          and ``j``.
        - 'lower': ``i``.
        - 'higher': ``j``.
        - 'nearest': ``i`` or ``j``, whichever is nearest.
        - 'midpoint': ``(i + j) / 2``.

        .. versionchanged:: 2022.1.0
            This argument was previously called "interpolation"

    internal_method : {'default', 'dask', 'tdigest'}, optional
        What internal method to use. By default will use dask's internal custom
        algorithm (``'dask'``).  If set to ``'tdigest'`` will use tdigest for
        floats and ints if method is linear and fallback to ``'dask'`` otherwise.

        .. versionchanged:: 2022.1.0
            This argument was previously called “method”.

    interpolation : str, optional
        Deprecated name for the method keyword argument.

        .. deprecated:: 2022.1.0

    See Also
    --------
    numpy.percentile : Numpy's equivalent Percentile function
    """

    if a.ndim == 1:
        from dask.array.utils import array_safe, meta_from_array

        allowed_internal_methods = {"default", "dask", "tdigest"}

        if method in allowed_internal_methods:
            warnings.warn(
                "The `method=` argument was renamed to `internal_method=`",
                FutureWarning,
            )
            internal_method = method

        if "interpolation" in kwargs:
            warnings.warn(
                "The `interpolation=` argument to percentile was renamed to `method= ` ",
                FutureWarning,
            )
            method = kwargs.pop("interpolation")

        if kwargs:
            raise TypeError(
                f"percentile() got an unexpected keyword argument {kwargs.keys()}"
            )

        q_is_number = False
        if isinstance(q, Number):
            q_is_number = True
            q = [q]
        q = array_safe(q, like=meta_from_array(a))
        token = tokenize(a, q, method)

        dtype = a.dtype
        if np.issubdtype(dtype, np.integer):
            dtype = (array_safe([], dtype=dtype, like=meta_from_array(a)) / 0.5).dtype
        meta = meta_from_array(a, dtype=dtype)

        if internal_method not in allowed_internal_methods:
            raise ValueError(
                f"`internal_method=` must be one of {allowed_internal_methods}"
            )

        # Allow using t-digest if method is allowed and dtype is of floating or integer type
        if (
            internal_method == "tdigest"
            and method == "linear"
            and (np.issubdtype(dtype, np.floating) or np.issubdtype(dtype, np.integer))
        ):
            from dask.utils import import_required

            import_required(
                "crick", "crick is a required dependency for using the t-digest method."
            )

            name = "percentile_tdigest_chunk-" + token
            dsk = {
                (name, i): (_tdigest_chunk, key)
                for i, key in enumerate(a.__dask_keys__())
            }

            name2 = "percentile_tdigest-" + token

            dsk2 = {(name2, 0): (_percentiles_from_tdigest, q, sorted(dsk))}

        # Otherwise use the custom percentile algorithm
        else:
            # Add 0 and 100 during calculation for more robust behavior (hopefully)
            zero = empty_lookup.dispatch(type(q))(1, dtype=q.dtype)
            zero[:] = 0

            hundred = empty_lookup.dispatch(type(q))(1, dtype=q.dtype)
            hundred[:] = 100

            calc_q = np.concatenate((zero, q, hundred))
            name = "percentile_chunk-" + token
            dsk = {
                (name, i): (percentile_lookup, key, calc_q, method)
                for i, key in enumerate(a.__dask_keys__())
            }

            name2 = "percentile-" + token
            dsk2 = {
                (name2, 0): (
                    merge_percentiles,
                    q,
                    [calc_q] * len(a.chunks[0]),
                    sorted(dsk),
                    method,
                )
            }
        dsk = merge(dsk, dsk2)
        graph = HighLevelGraph.from_collections(name2, dsk, dependencies=[a])
        arr = Array(graph, name2, chunks=((len(q),),), meta=meta)
        return arr.reshape(()) if q_is_number else arr

    elif a.ndim > 1:
        from dask.array.reductions import quantile

        q = np.true_divide(q, a.dtype.type(100) if a.dtype.kind == "f" else 100)
        return quantile(a, q, method=method, **kwargs)
    else:
        raise NotImplementedError("support for arrays of ndim 0 is not implemented.")


def merge_percentiles(finalq, qs, vals, method="lower", Ns=None, raise_on_nan=True):
    """Combine several percentile calculations of different data.

    Parameters
    ----------

    finalq : numpy.array
        Percentiles to compute (must use same scale as ``qs``).
    qs : sequence of :class:`numpy.array`s
        Percentiles calculated on different sets of data.
    vals : sequence of :class:`numpy.array`s
        Resulting values associated with percentiles ``qs``.
    Ns : sequence of integers
        The number of data elements associated with each data set.
    method : {'linear', 'lower', 'higher', 'midpoint', 'nearest'}
        Specify the interpolation method to use to calculate final
        percentiles.  For more information, see :func:`numpy.percentile`.

    Examples
    --------

    >>> finalq = [10, 20, 30, 40, 50, 60, 70, 80]
    >>> qs = [[20, 40, 60, 80], [20, 40, 60, 80]]
    >>> vals = [np.array([1, 2, 3, 4]), np.array([10, 11, 12, 13])]
    >>> Ns = [100, 100]  # Both original arrays had 100 elements

    >>> merge_percentiles(finalq, qs, vals, Ns=Ns)
    array([ 1,  2,  3,  4, 10, 11, 12, 13])
    """
    from dask.array.utils import array_safe

    if isinstance(finalq, Iterator):
        finalq = list(finalq)
    finalq = array_safe(finalq, like=finalq)
    qs = [list(q) for q in qs]
    vals = list(vals)
    if Ns is None:
        vals, Ns = zip(*vals)
    Ns = list(Ns)

    L = list(zip(*((q, val, N) for q, val, N in zip(qs, vals, Ns) if N)))
    if not L:
        if raise_on_nan:
            raise ValueError("No non-trivial arrays found")
        return np.full(len(qs[0]) - 2, np.nan)
    qs, vals, Ns = L

    # TODO: Perform this check above in percentile once dtype checking is easy
    #       Here we silently change meaning
    if vals[0].dtype.name == "category":
        result = merge_percentiles(
            finalq, qs, [v.codes for v in vals], method, Ns, raise_on_nan
        )
        import pandas as pd

        return pd.Categorical.from_codes(result, vals[0].categories, vals[0].ordered)
    if not np.issubdtype(vals[0].dtype, np.number):
        method = "nearest"

    if len(vals) != len(qs) or len(Ns) != len(qs):
        raise ValueError("qs, vals, and Ns parameters must be the same length")

    # transform qs and Ns into number of observations between percentiles
    total_len = sum(len(q) for q in qs)
    counts = empty_lookup.dispatch(type(finalq))(total_len, dtype=finalq.dtype)
    start = 0
    for q, N in zip(qs, Ns):
        length = len(q)
        count = empty_lookup.dispatch(type(finalq))(length, dtype=finalq.dtype)
        count[1:] = np.diff(array_safe(q, like=q[0]))
        count[0] = q[0]
        count *= N
        counts[start : start + length] = count
        start += length

    # Sort by calculated percentile values, then number of observations.
    combined_vals = np.concatenate(vals)
    combined_counts = array_safe(counts, like=combined_vals)
    sort_order = np.argsort(combined_vals)
    combined_vals = np.take(combined_vals, sort_order)
    combined_counts = np.take(combined_counts, sort_order)

    # percentile-like, but scaled by total number of observations
    combined_q = np.cumsum(combined_counts)

    # rescale finalq percentiles to match combined_q
    finalq = array_safe(finalq, like=combined_vals)
    desired_q = finalq * sum(Ns)

    # the behavior of different interpolation methods should be
    # investigated further.
    if method == "linear":
        rv = np.interp(desired_q, combined_q, combined_vals)
    else:
        left = np.searchsorted(combined_q, desired_q, side="left")
        right = np.searchsorted(combined_q, desired_q, side="right") - 1
        np.minimum(left, len(combined_vals) - 1, out=left)  # don't exceed max index
        lower = np.minimum(left, right)
        upper = np.maximum(left, right)
        if method == "lower":
            rv = combined_vals[lower]
        elif method == "higher":
            rv = combined_vals[upper]
        elif method == "midpoint":
            rv = 0.5 * (combined_vals[lower] + combined_vals[upper])
        elif method == "nearest":
            lower_residual = np.abs(combined_q[lower] - desired_q)
            upper_residual = np.abs(combined_q[upper] - desired_q)
            mask = lower_residual > upper_residual
            index = lower  # alias; we no longer need lower
            index[mask] = upper[mask]
            rv = combined_vals[index]
        else:
            raise ValueError(
                "interpolation method can only be 'linear', 'lower', "
                "'higher', 'midpoint', or 'nearest'"
            )
    return rv


@derived_from(np)
def nanpercentile(a, q, **kwargs):
    from dask.array.reductions import nanquantile

    q = np.true_divide(q, a.dtype.type(100) if a.dtype.kind == "f" else 100)

    return nanquantile(a, q, **kwargs)



import pytest

import dask.config
from dask.array import _array_expr_enabled, array_expr_enabled

pytestmark = pytest.mark.normal_and_array_expr


def test_array_expr_enabled():
    assert _array_expr_enabled() in (True, False)
    assert array_expr_enabled() is _array_expr_enabled()


def test_flip_config():
    """Config changes after the initial import are ignored with a warning"""
    prev = _array_expr_enabled()
    with dask.config.set({"array.query-planning": not prev}):
        with pytest.warns(RuntimeWarning, match="first imported"):
            actual = _array_expr_enabled()
    assert actual == prev


def test_str_config():
    """Prevent accidental use of 'false' string in YAML config,
    which would evaluate to True.
    """
    with dask.config.set({"array.query-planning": "false"}):
        with pytest.raises(TypeError):
            _array_expr_enabled()



from __future__ import annotations

import math
import os
import sys

try:
    import psutil
except ImportError:
    psutil = None  # type: ignore[assignment]

from dask._compatibility import LINUX

__all__ = ("cpu_count", "CPU_COUNT")


def _try_extract_cgroup_cpu_quota():
    # cgroup v1
    # The directory name isn't standardized across linux distros, check both
    for dirname in ["cpuacct,cpu", "cpu,cpuacct"]:
        try:
            with open(f"/sys/fs/cgroup/{dirname}/cpu.cfs_quota_us") as f:
                quota = int(f.read())
            with open(f"/sys/fs/cgroup/{dirname}/cpu.cfs_period_us") as f:
                period = int(f.read())
            return quota, period
        except Exception:
            pass

    # cgroup v2
    try:
        with open("/proc/self/cgroup") as f:
            group_path = f.read().strip().split(":")[-1]
        if not group_path.endswith("/"):
            group_path = f"{group_path}/"
        with open(f"/sys/fs/cgroup{group_path}cpu.max") as f:
            quota, period = map(int, f.read().split(" "))
            return quota, period
    except Exception:
        pass

    # No cgroup CPU quota found
    return None, None


def cpu_count() -> int:
    """Get the available CPU count for this system.

    Takes the minimum value from the following locations:

    - Total system cpus available on the host.
    - CPU Affinity (if set)
    - Cgroups limit (if set)
    """
    if sys.version_info >= (3, 13):
        # Embeds CPU affinity checks
        count = os.process_cpu_count()
    elif hasattr(os, "sched_getaffinity"):
        # https://docs.python.org/3/library/os.html#interface-to-the-scheduler
        # "only available on some Unix platforms"; neither MacOS nor Windows
        count = len(os.sched_getaffinity(0))
    else:
        # Does not account for CPU affinity.
        # On exotic alternative Python implementations, it may return None.
        count = os.cpu_count() or 1
    assert count

    # Additional CPU affinity check with psutil.
    # NOTE: do not limit this to Python <3.13: on Windows,
    # `psutil.Process().cpu_affinity(value)` does not change the reading of
    # os.process_cpu_count().
    if psutil is not None:
        proc = psutil.Process()
        if hasattr(proc, "cpu_affinity"):
            affinity = proc.cpu_affinity()
            if affinity is not None:
                assert affinity
                count = min(count, len(affinity))

    # Check cgroups if available
    if LINUX:
        quota, period = _try_extract_cgroup_cpu_quota()
        if quota is not None and period is not None:
            # We round up on fractional CPUs
            cgroups_count = math.ceil(quota / period)
            if cgroups_count > 0:
                count = min(count, cgroups_count)

    return count


CPU_COUNT = cpu_count()



from __future__ import annotations

import builtins
import io
import os
import subprocess
import sys
from contextlib import suppress

import pytest

from dask._compatibility import LINUX, MACOS
from dask.system import CPU_COUNT, cpu_count


def test_cpu_count():
    count = cpu_count()
    assert isinstance(count, int)
    assert count == CPU_COUNT
    assert count <= (os.cpu_count() or 999)
    assert count >= 1


@pytest.mark.skipif(MACOS, reason="No CPU affinity in psutil")
@pytest.mark.parametrize(("affinity"), [{0}, {1}, {0, 1}, {0, 2}])
def test_cpu_affinity_psutil(affinity):
    """Test that cpu_count() respects CPU affinity set by psutil"""
    psutil = pytest.importorskip("psutil")
    proc = psutil.Process()
    prev = proc.cpu_affinity()
    if prev is None:
        pytest.skip("No support for CPU affinity")
    if not affinity.issubset(set(prev)):
        pytest.skip("Not enough CPUs")  # pragma: no cover

    proc.cpu_affinity(affinity)
    try:
        assert cpu_count() == len(affinity)
    finally:
        proc.cpu_affinity(prev)


@pytest.mark.skipif(not LINUX, reason="No command line API for CPU affinity")
@pytest.mark.parametrize(("affinity"), [{0}, {1}, {0, 1}, {0, 2}])
def test_cpu_affinity_taskset(affinity):
    """Test that cpu_count() respects the taskset command line tool"""
    count = len(affinity)
    if (os.cpu_count() or 1) < count:
        raise pytest.skip("Not enough CPUs")  # pragma: no cover

    subprocess.check_call(
        [
            "taskset",
            "-c",
            ",".join(str(i) for i in sorted(affinity)),
            sys.executable,
            "-c",
            f"from dask.system import CPU_COUNT; assert CPU_COUNT == {count}",
        ]
    )


@pytest.mark.skipif(
    sys.implementation.name != "cpython" or sys.version_info < (3, 13),
    reason="-X cpu_count= added in CPython 3.13",
)
@pytest.mark.parametrize("count", [1, 2, 3])
def test_cpu_count_arg(count):
    """Test that cpu_count() respects the python -X cpu_count= parameter"""
    if (os.cpu_count() or 1) < count:
        raise pytest.skip("Not enough CPUs")  # pragma: no cover

    subprocess.check_call(
        [
            sys.executable,
            "-X",
            f"cpu_count={count}",
            "-c",
            f"from dask.system import CPU_COUNT; assert CPU_COUNT == {count}",
        ]
    )


@pytest.fixture
def monkeypatch_cpu_count(monkeypatch):
    def cpu_count():
        return 250  # Absurdly high, unlikely to match real value

    def sched_getaffinity(pid):
        return set(range(250))

    class Process:
        def cpu_affinity(self):
            return list(range(250))

    monkeypatch.setattr(os, "cpu_count", cpu_count)
    if sys.version_info >= (3, 13):
        monkeypatch.setattr(os, "process_cpu_count", cpu_count)
    monkeypatch.setattr(os, "sched_getaffinity", sched_getaffinity)
    with suppress(ImportError):
        import psutil

        monkeypatch.setattr(psutil, "Process", Process)


@pytest.mark.skipif(not LINUX, reason="Control Groups only available on Linux")
@pytest.mark.parametrize("dirname", ["cpuacct,cpu", "cpu,cpuacct", None])
def test_cpu_count_cgroups(dirname, monkeypatch, monkeypatch_cpu_count):
    if dirname:
        paths = {
            f"/sys/fs/cgroup/{dirname}/cpu.cfs_quota_us": io.StringIO("2005"),
            f"/sys/fs/cgroup/{dirname}/cpu.cfs_period_us": io.StringIO("10"),
        }
        builtin_open = builtins.open

        def myopen(path, *args, **kwargs):
            if path in paths:
                return paths.get(path)
            return builtin_open(path, *args, **kwargs)

        monkeypatch.setattr(builtins, "open", myopen)
        monkeypatch.setattr(sys, "platform", "linux")

    count = cpu_count()
    if dirname:
        # Rounds up
        assert count == 201
    else:
        assert count == 250


@pytest.mark.skipif(not LINUX, reason="Control Groups only available on Linux")
@pytest.mark.parametrize("group_name", ["/", "/user.slice", "/user.slice/more.slice"])
@pytest.mark.parametrize("quota", ["max", "2005"])
def test_cpu_count_cgroups_v2(quota, group_name, monkeypatch, monkeypatch_cpu_count):
    if not group_name.endswith("/"):
        group_name = f"{group_name}/"

    paths = {
        "/proc/self/cgroup": io.StringIO(f"0::{group_name}"),
        f"/sys/fs/cgroup{group_name}cpu.max": io.StringIO(f"{quota} 10"),
    }
    builtin_open = builtins.open

    def myopen(path, *args, **kwargs):
        if path in paths:
            return paths.get(path)
        return builtin_open(path, *args, **kwargs)

    monkeypatch.setattr(builtins, "open", myopen)
    monkeypatch.setattr(sys, "platform", "linux")

    count = cpu_count()
    if quota == "max":
        assert count == 250
    else:
        # Rounds up
        assert count == 201



from __future__ import annotations

import contextvars
import signal
import sys
import threading
import warnings
from concurrent.futures import ThreadPoolExecutor
from multiprocessing.pool import ThreadPool
from time import sleep, time

import pytest

import dask
from dask.system import CPU_COUNT
from dask.threaded import get
from dask.utils_test import add, inc


def test_get():
    dsk = {"x": 1, "y": 2, "z": (inc, "x"), "w": (add, "z", "y")}
    assert get(dsk, "w") == 4
    assert get(dsk, ["w", "z"]) == (4, 2)


def test_nested_get():
    dsk = {"x": 1, "y": 2, "a": (add, "x", "y"), "b": (sum, ["x", "y"])}
    assert get(dsk, ["a", "b"]) == (3, 3)


def test_get_without_computation():
    dsk = {"x": 1}
    assert get(dsk, "x") == 1


def test_broken_callback():
    from dask.callbacks import Callback

    def _f_ok(*args, **kwargs):
        pass

    def _f_broken(*args, **kwargs):
        raise ValueError("my_exception")

    dsk = {"x": 1}

    with Callback(start=_f_broken, finish=_f_ok):
        with Callback(start=_f_ok, finish=_f_ok):
            with pytest.raises(ValueError, match="my_exception"):
                get(dsk, "x")


def bad(x):
    raise ValueError()


def test_exceptions_rise_to_top():
    dsk = {"x": 1, "y": (bad, "x")}
    pytest.raises(ValueError, lambda: get(dsk, "y"))


@pytest.mark.parametrize("pool_typ", [ThreadPool, ThreadPoolExecutor])
def test_reuse_pool(pool_typ):
    with pool_typ(CPU_COUNT) as pool:
        with dask.config.set(pool=pool):
            assert get({"x": (inc, 1)}, "x") == 2
            assert get({"x": (inc, 1)}, "x") == 2


@pytest.mark.parametrize("pool_typ", [ThreadPool, ThreadPoolExecutor])
def test_pool_kwarg(pool_typ):
    def f():
        sleep(0.01)
        return threading.get_ident()

    dsk = {("x", i): (f,) for i in range(30)}
    dsk["x"] = (len, (set, [("x", i) for i in range(len(dsk))]))

    with pool_typ(3) as pool:
        assert get(dsk, "x", pool=pool) == 3


def test_threaded_within_thread():
    L = []

    def f(i):
        result = get({"x": (lambda: i,)}, "x", num_workers=2)
        L.append(result)

    before = threading.active_count()

    for _ in range(20):
        t = threading.Thread(target=f, args=(1,))
        t.daemon = True
        t.start()
        t.join()
        assert L == [1]
        del L[:]

    start = time()  # wait for most threads to join
    while threading.active_count() > before + 10:
        sleep(0.01)
        assert time() < start + 5


def test_dont_spawn_too_many_threads():
    before = threading.active_count()

    dsk = {("x", i): (lambda i=i: i,) for i in range(10)}
    dsk["x"] = (sum, list(dsk))
    for _ in range(20):
        get(dsk, "x", num_workers=4)

    after = threading.active_count()

    assert after <= before + 8


def test_dont_spawn_too_many_threads_CPU_COUNT():
    before = threading.active_count()

    dsk = {("x", i): (lambda i=i: i,) for i in range(10)}
    dsk["x"] = (sum, list(dsk))
    for _ in range(20):
        get(dsk, "x")

    after = threading.active_count()

    assert after <= before + CPU_COUNT * 2


def test_thread_safety():
    def f(x):
        return 1

    dsk = {"x": (sleep, 0.05), "y": (f, "x")}

    L = []

    def test_f():
        L.append(get(dsk, "y"))

    threads = []
    for _ in range(20):
        t = threading.Thread(target=test_f)
        t.daemon = True
        t.start()
        threads.append(t)

    for thread in threads:
        thread.join()

    assert L == [1] * 20


def test_interrupt():
    # Windows implements `queue.get` using polling,
    # which means we can set an exception to interrupt the call to `get`.
    # Python 3 on other platforms requires sending SIGINT to the main thread.
    if sys.platform == "win32":
        from _thread import interrupt_main
    else:
        main_thread = threading.get_ident()

        def interrupt_main() -> None:
            signal.pthread_kill(main_thread, signal.SIGINT)

    in_clog_event = threading.Event()
    clog_event = threading.Event()

    def clog(in_clog_event: threading.Event, clog_event: threading.Event) -> None:
        in_clog_event.set()
        clog_event.wait()

    def interrupt(in_clog_event: threading.Event) -> None:
        in_clog_event.wait()
        interrupt_main()

    dsk = {("x", i): (clog, in_clog_event, clog_event) for i in range(20)}
    dsk["x"] = (len, list(dsk.keys()))

    interrupter = threading.Thread(target=interrupt, args=(in_clog_event,))
    interrupter.start()

    # Use explicitly created ThreadPoolExecutor to avoid leaking threads after
    # the KeyboardInterrupt
    with ThreadPoolExecutor(CPU_COUNT) as pool:
        with pytest.raises(KeyboardInterrupt):
            get(dsk, "x", pool=pool)
        clog_event.set()
    interrupter.join()


def warm_up_thread_pool(num_workers: int | None) -> None:
    """Ensure that there is a warm thread pool.

    With num_workers=None on the default thread, this function warms up the default
    thread pool. With num_workers set or from a different thread, this function spawns
    and warms up a non-default thread pool.

    When this function is called after another test that called get() or compute(), the
    thread pool will be already warm, at least partially (not all threads will
    necessarily be already running).
    """
    parties = num_workers or CPU_COUNT
    barrier = threading.Barrier(parties)
    dsk = {("x", i): (barrier.wait,) for i in range(parties)}
    get(dsk, list(dsk), num_workers=num_workers)


test_ctxvar = contextvars.ContextVar("test_ctxvar", default=42)


@pytest.mark.parametrize("num_workers", [None, 1])
def test_contextvars(num_workers):
    """When the thread pool is cold, contextvars area automatically propagated from the
    caller thread to **new** threads on Python 3.14t by default, as well as on 3.14 with
    the opt-in flag `python -X thread_inherit_context=1`.

    Test that contextvars are always propagated to the tasks, even in absence of the above
    flag and even when the threads were already warm when the context was set.
    """
    warm_up_thread_pool(num_workers)

    dsk = {"x": (test_ctxvar.get,)}
    # Reuse thread pool started by warm_up_thread_pool
    assert get(dsk, "x", num_workers=num_workers) == 42
    tok = test_ctxvar.set(43)
    try:
        assert get(dsk, "x", num_workers=num_workers) == 43
    finally:
        test_ctxvar.reset(tok)


@pytest.mark.parametrize("num_workers", [None, 1])
def test_context_aware_warnings(num_workers):
    """Test that warnings issued from a thread pool can be caught by the main thread,
    even when the thread pool was already warm when entering `warnings.catch_warnings`.
    This also affects `@pytest.mark.filterwarnings` and similar mechanisms.

    This needs special handling on Python 3.14t, which uses context-aware warnings.
    The system is also available opt-in on 3.14 (GIL-enabled) with
       python -X context_aware_warnings=1
    """
    warm_up_thread_pool(num_workers)

    def warn():
        warnings.warn("Hello world", RuntimeWarning, stacklevel=2)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        # Reuse thread pool started by warm_up_thread_pool
        get({"x": (warn,)}, "x", num_workers=num_workers)

    assert len(w) == 1
    assert isinstance(w[0].message, RuntimeWarning)
    assert str(w[0].message) == "Hello world"



from __future__ import annotations

from dask import config, datasets
from dask._expr import Expr, HLGExpr, LLGExpr, SingletonExpr

try:
    # Backwards compatibility with versioneer
    from dask._version import __commit_id__ as __git_revision__
    from dask._version import __version__
except ImportError:  # pragma: no cover
    __git_revision__ = "unknown"
    __version__ = "unknown"
from dask.base import (
    annotate,
    compute,
    get_annotations,
    is_dask_collection,
    optimize,
    persist,
    visualize,
)
from dask.core import istask
from dask.delayed import delayed
from dask.local import get_sync as get



from __future__ import annotations

import functools
from collections import OrderedDict, UserDict
from collections.abc import Hashable
from typing import Any, Literal, TypeVar, cast

import pandas as pd

from dask import config, is_dask_collection
from dask.dataframe._compat import is_string_dtype
from dask.dataframe.core import is_dataframe_like, is_series_like
from dask.tokenize import _tokenize_deterministic, normalize_token
from dask.utils import get_default_shuffle_method

K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def _calc_maybe_new_divisions(df, periods, freq):
    """Maybe calculate new divisions by periods of size freq

    Used to shift the divisions for the `shift` method. If freq isn't a fixed
    size (not anchored or relative), then the divisions are shifted
    appropriately.

    Returning None, indicates divisions ought to be cleared.

    Parameters
    ----------
    df : dd.DataFrame, dd.Series, or dd.Index
    periods : int
        The number of periods to shift.
    freq : DateOffset, timedelta, or time rule string
        The frequency to shift by.
    """
    if isinstance(freq, str):
        freq = pd.tseries.frequencies.to_offset(freq)

    is_offset = isinstance(freq, pd.DateOffset)
    if is_offset:
        if not isinstance(freq, (pd.offsets.Tick, pd.offsets.Day)):
            # Can't infer divisions on relative or anchored offsets, as
            # divisions may now split identical index value.
            # (e.g. index_partitions = [[1, 2, 3], [3, 4, 5]])
            # As of https://github.com/pandas-dev/pandas/pull/61985
            # Day doesn't subclass Tick.
            return None  # Would need to clear divisions
    if df.known_divisions:
        divs = pd.Series(range(len(df.divisions)), index=df.divisions)
        divisions = divs.shift(periods, freq=freq).index
        return tuple(divisions)
    return df.divisions


def _validate_axis(axis=0, none_is_zero: bool = True) -> None | Literal[0, 1]:
    if axis not in (0, 1, "index", "columns", None):
        raise ValueError(f"No axis named {axis}")
    # convert to numeric axis
    numeric_axis: dict[str | None, Literal[0, 1]] = {"index": 0, "columns": 1}
    if none_is_zero:
        numeric_axis[None] = 0

    return numeric_axis.get(axis, axis)


def _convert_to_list(column) -> list | None:
    if column is None or isinstance(column, list):
        pass
    elif isinstance(column, tuple):
        column = list(column)
    elif hasattr(column, "dtype"):
        column = column.tolist()
    else:
        column = [column]
    return column


def _tokenize_partial(expr, ignore: list | None = None) -> str:
    # Helper function to "tokenize" the operands
    # that are not in the `ignore` list
    ignore = ignore or []
    return _tokenize_deterministic(
        *[
            op
            for i, op in enumerate(expr.operands)
            if i >= len(expr._parameters) or expr._parameters[i] not in ignore
        ]
    )


class LRU(UserDict[K, V]):
    """Limited size mapping, evicting the least recently looked-up key when full"""

    def __init__(self, maxsize: float) -> None:
        super().__init__()
        self.data = OrderedDict()
        self.maxsize = maxsize

    def __getitem__(self, key: K) -> V:
        value = super().__getitem__(key)
        cast(OrderedDict, self.data).move_to_end(key)
        return value

    def __setitem__(self, key: K, value: V) -> None:
        if len(self) >= self.maxsize:
            cast(OrderedDict, self.data).popitem(last=False)
        super().__setitem__(key, value)


class _BackendData:
    """Helper class to wrap backend data

    The primary purpose of this class is to provide
    caching outside the ``FromPandas`` class.
    """

    def __init__(self, data):
        self._data = data
        self._division_info = LRU(10)

    @functools.cached_property
    def _token(self):
        from dask.tokenize import _tokenize_deterministic

        return _tokenize_deterministic(self._data)

    def __len__(self):
        return len(self._data)

    def __getattr__(self, key: str) -> Any:
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            # Return the underlying backend attribute
            return getattr(self._data, key)

    def __reduce__(self):
        return type(self), (self._data,)

    def __deepcopy__(self, memodict=None):
        return type(self)(self._data.copy())


@normalize_token.register(_BackendData)
def normalize_data_wrapper(data):
    return data._token


def _maybe_from_pandas(dfs):
    from dask.dataframe.dask_expr import from_pandas

    def _pd_series_or_dataframe(x):
        # `x` can be a cudf Series/DataFrame
        return not is_dask_collection(x) and (is_series_like(x) or is_dataframe_like(x))

    dfs = [from_pandas(df, 1) if _pd_series_or_dataframe(df) else df for df in dfs]
    return dfs


def _get_shuffle_preferring_order(shuffle):
    if shuffle is not None:
        return shuffle

    # Choose tasks over disk since it keeps the order
    shuffle = get_default_shuffle_method()
    if shuffle == "disk":
        return "tasks"

    return shuffle


def _raise_if_object_series(x, funcname):
    """
    Utility function to raise an error if an object column does not support
    a certain operation like `mean`.
    """
    if x.ndim == 1 and hasattr(x, "dtype"):
        if x.dtype == object:
            raise ValueError(f"`{funcname}` not supported with object series")
        elif is_string_dtype(x):
            raise ValueError(f"`{funcname}` not supported with string series")


def _is_any_real_numeric_dtype(arr_or_dtype):
    try:
        from pandas.api.types import is_any_real_numeric_dtype

        return is_any_real_numeric_dtype(arr_or_dtype)
    except ImportError:
        # Temporary/soft pandas<2 support to enable cudf dev
        # TODO: Remove `try` block after 4/2024
        from pandas.api.types import is_bool_dtype, is_complex_dtype, is_numeric_dtype

        return (
            is_numeric_dtype(arr_or_dtype)
            and not is_complex_dtype(arr_or_dtype)
            and not is_bool_dtype(arr_or_dtype)
        )


def get_specified_shuffle(shuffle_method):
    # Take the config shuffle if given, otherwise defer evaluation until optimize
    return shuffle_method or config.get("dataframe.shuffle.method", None)



from __future__ import annotations

import os
import pathlib
import random
import site
import stat
import sys
from collections import OrderedDict
from contextlib import contextmanager

import pytest
import yaml

import dask.config
from dask.config import (
    _get_paths,
    canonical_name,
    collect,
    collect_env,
    collect_yaml,
    config,
    deserialize,
    ensure_file,
    expand_environment_variables,
    get,
    merge,
    paths_containing_key,
    pop,
    refresh,
    rename,
    serialize,
    update,
    update_defaults,
)


def test_canonical_name():
    c = {"foo-bar": 1, "fizz_buzz": 2}
    assert canonical_name("foo-bar", c) == "foo-bar"
    assert canonical_name("foo_bar", c) == "foo-bar"
    assert canonical_name("fizz-buzz", c) == "fizz_buzz"
    assert canonical_name("fizz_buzz", c) == "fizz_buzz"
    assert canonical_name("new-key", c) == "new-key"
    assert canonical_name("new_key", c) == "new_key"


def test_update():
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": OrderedDict({"b": 2})}
    update(b, a)
    assert b == {"x": 1, "y": {"a": 1, "b": 2}, "z": 3}

    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": {"a": 3, "b": 2}}
    update(b, a, priority="old")
    assert b == {"x": 2, "y": {"a": 3, "b": 2}, "z": 3}


def test_update_new_defaults():
    d = {"x": 1, "y": 1, "z": {"a": 1, "b": 1}}
    o = {"x": 1, "y": 2, "z": {"a": 1, "b": 2}, "c": 2, "c2": {"d": 2}}
    n = {"x": 3, "y": 3, "z": OrderedDict({"a": 3, "b": 3}), "c": 3, "c2": {"d": 3}}
    assert update(o, n, priority="new-defaults", defaults=d) == {
        "x": 3,
        "y": 2,
        "z": {"a": 3, "b": 2},
        "c": 2,
        "c2": {"d": 2},
    }
    assert update(o, n, priority="new-defaults", defaults=o) == update(
        o, n, priority="new"
    )
    assert update(o, n, priority="new-defaults", defaults=None) == update(
        o, n, priority="old"
    )


def test_update_defaults():
    defaults = [
        {"a": 1, "b": {"c": 1}},
        {"a": 2, "b": {"d": 2}},
    ]
    current = {"a": 2, "b": {"c": 1, "d": 3}, "extra": 0}
    new = {"a": 0, "b": {"c": 0, "d": 0}, "new-extra": 0}
    update_defaults(new, current, defaults=defaults)

    assert defaults == [
        {"a": 1, "b": {"c": 1}},
        {"a": 2, "b": {"d": 2}},
        {"a": 0, "b": {"c": 0, "d": 0}, "new-extra": 0},
    ]
    assert current == {"a": 0, "b": {"c": 0, "d": 3}, "extra": 0, "new-extra": 0}


def test_update_list_to_dict():
    a = {"x": {"y": 1, "z": 2}}
    b = {"x": [1, 2], "w": 3}
    update(b, a)
    assert b == {"x": {"y": 1, "z": 2}, "w": 3}


def test_update_dict_to_list():
    a = {"x": [1, 2]}
    b = {"x": {"y": 1, "z": 2}, "w": 3}
    update(b, a)
    assert b == {"x": [1, 2], "w": 3}


def test_merge():
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": {"b": 2}}

    expected = {"x": 2, "y": {"a": 1, "b": 2}, "z": 3}

    c = merge(a, b)
    assert c == expected


def test_collect_yaml_paths(tmp_path):
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": {"b": 2}}

    expected = {"x": 2, "y": {"a": 1, "b": 2}, "z": 3}

    pa, pb = tmp_path / "a.yaml", tmp_path / "b.yaml"
    pa.write_text(yaml.dump(a))
    pb.write_text(yaml.dump(b))

    configs = list(collect_yaml(paths=[pa, pb], return_paths=True))
    assert configs == [(pa, a), (pb, b)]

    # `return_paths` defaults to False
    config = merge(*collect_yaml(paths=[pa, pb]))
    assert config == expected


def test_paths_containing_key(tmp_path):
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": {"b": 2}}

    pa, pb = tmp_path / "a.yaml", tmp_path / "b.yaml"
    pa.write_text(yaml.dump(a))
    pb.write_text(yaml.dump(b))

    paths = list(paths_containing_key("y.a", paths=[pa, pb]))
    assert paths == [pa]

    assert not list(paths_containing_key("w", paths=[pa, pb]))
    assert not list(paths_containing_key("x", paths=[]))


def test_collect_yaml_dir(tmp_path):
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": {"b": 2}}

    expected = {"x": 2, "y": {"a": 1, "b": 2}, "z": 3}

    pa, pb = tmp_path / "a.yaml", tmp_path / "b.yaml"
    pa.write_text(yaml.dump(a))
    pb.write_text(yaml.dump(b))

    config = merge(*collect_yaml(paths=[tmp_path]))
    assert config == expected


@contextmanager
def no_read_permissions(path):
    perm_orig = stat.S_IMODE(os.stat(path).st_mode)
    perm_new = perm_orig ^ stat.S_IREAD
    try:
        os.chmod(path, perm_new)
        yield
    finally:
        os.chmod(path, perm_orig)


@pytest.mark.skipif(
    sys.platform == "win32", reason="Can't make writeonly file on windows"
)
@pytest.mark.parametrize("kind", ["directory", "file"])
def test_collect_yaml_permission_errors(tmp_path, kind):
    a = {"x": 1, "y": 2}
    b = {"y": 3, "z": 4}

    pa, pb = tmp_path / "a.yaml", tmp_path / "b.yaml"
    pa.write_text(yaml.dump(a))
    pb.write_text(yaml.dump(b))

    if kind == "directory":
        cant_read = tmp_path
        expected = {}
    else:
        cant_read = pa
        expected = b

    with no_read_permissions(cant_read):
        config = merge(*collect_yaml(paths=[tmp_path]))
        assert config == expected


def test_collect_yaml_malformed_file(tmp_path):
    (tmp_path / "a.yaml").write_bytes(b"{")

    with pytest.raises(ValueError) as rec:
        list(collect_yaml(paths=[tmp_path]))
    assert "a.yaml" in str(rec.value)
    assert "is malformed" in str(rec.value)
    assert "original error message" in str(rec.value)


def test_collect_yaml_no_top_level_dict(tmp_path):
    (tmp_path / "a.yaml").write_bytes(b"[1234]")

    with pytest.raises(ValueError) as rec:
        list(collect_yaml(paths=[tmp_path]))
    assert "a.yaml" in str(rec.value)
    assert "is malformed" in str(rec.value)
    assert "must have a dict" in str(rec.value)


def test_env():
    env = {
        "DASK_A_B": "123",
        "DASK_C": "True",
        "DASK_D": "hello",
        "DASK_E__X": "123",
        "DASK_E__Y": "456",
        "DASK_F": '[1, 2, "3"]',
        "DASK_G": "/not/parsable/as/literal",
        "FOO": "not included",
    }

    expected = {
        "a_b": 123,
        "c": True,
        "d": "hello",
        "e": {"x": 123, "y": 456},
        "f": [1, 2, "3"],
        "g": "/not/parsable/as/literal",
    }

    res = collect_env(env)
    assert res == expected


@pytest.mark.parametrize(
    "preproc", [lambda x: x, lambda x: x.lower(), lambda x: x.upper()]
)
@pytest.mark.parametrize(
    "v,out", [("None", None), ("Null", None), ("False", False), ("True", True)]
)
def test_env_special_values(preproc, v, out):
    env = {"DASK_A": preproc(v)}
    res = collect_env(env)
    assert res == {"a": out}


def test_collect(tmp_path):
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 2, "z": 3, "y": {"b": 2}}
    env = {"DASK_W": "4"}

    expected = {"w": 4, "x": 2, "y": {"a": 1, "b": 2}, "z": 3}

    pa, pb = tmp_path / "a.yaml", tmp_path / "b.yaml"
    pa.write_text(yaml.dump(a))
    pb.write_text(yaml.dump(b))

    config = collect([pa, pb], env=env)
    assert config == expected


def test_collect_env_none(monkeypatch):
    monkeypatch.setenv("DASK_FOO", "bar")
    config = collect([])
    assert config.get("foo") == "bar"


def test_get():
    d = {"x": 1, "y": {"a": 2}}

    assert get("x", config=d) == 1
    assert get("y.a", config=d) == 2
    assert get("y.b", 123, config=d) == 123
    with pytest.raises(KeyError):
        get("y.b", config=d)


def test_ensure_file(tmp_path):
    a = {"x": 1, "y": {"a": 1}}
    b = {"x": 123}

    source = tmp_path / "source.yaml"
    dest = tmp_path / "dest"
    destination = tmp_path / "dest" / "source.yaml"
    source.write_text(yaml.dump(a))

    ensure_file(source=source, destination=dest, comment=False)

    result = yaml.safe_load(destination.read_text())
    assert result == a

    # don't overwrite old config files
    source.write_text(yaml.dump(b))

    ensure_file(source=source, destination=dest, comment=False)

    result = yaml.safe_load(destination.read_text())
    assert result == a

    os.remove(destination)

    # Write again, now with comments
    ensure_file(source=source, destination=dest, comment=True)

    text = destination.read_text()
    assert "123" in text

    result = yaml.safe_load(destination.read_text())
    assert not result


def test_set():
    with dask.config.set(abc=123):
        assert config["abc"] == 123
        with dask.config.set(abc=456):
            assert config["abc"] == 456
        assert config["abc"] == 123

    assert "abc" not in config

    with dask.config.set({"abc": 123}):
        assert config["abc"] == 123
    assert "abc" not in config

    with dask.config.set({"abc.x": 1, "abc.y": 2, "abc.z.a": 3}):
        assert config["abc"] == {"x": 1, "y": 2, "z": {"a": 3}}
    assert "abc" not in config

    d = {}
    dask.config.set({"abc.x": 123}, config=d)
    assert d["abc"]["x"] == 123

    # Respect previous hyphenation, if any, or new hyphenation, if previous is not found
    d = {"e_f": 0, "g-h": 1}
    dask.config.set({"a_b": 2, "c-d": 3, "e-f": 4, "g_h": 5}, config=d)
    assert d == {"a_b": 2, "c-d": 3, "e_f": 4, "g-h": 5}


def test_set_kwargs():
    with dask.config.set(foo__bar=1, foo__baz=2):
        assert config["foo"] == {"bar": 1, "baz": 2}
    assert "foo" not in config

    # Mix kwargs and dict, kwargs override
    with dask.config.set({"foo.bar": 1, "foo.baz": 2}, foo__buzz=3, foo__bar=4):
        assert config["foo"] == {"bar": 4, "baz": 2, "buzz": 3}
    assert "foo" not in config

    # Mix kwargs and nested dict, kwargs override
    with dask.config.set({"foo": {"bar": 1, "baz": 2}}, foo__buzz=3, foo__bar=4):
        assert config["foo"] == {"bar": 4, "baz": 2, "buzz": 3}
    assert "foo" not in config


def test_set_nested():
    with dask.config.set({"abc": {"x": 123}}):
        assert config["abc"] == {"x": 123}
        with dask.config.set({"abc.y": 456}):
            assert config["abc"] == {"x": 123, "y": 456}
        assert config["abc"] == {"x": 123}
    assert "abc" not in config


def test_set_hard_to_copyables():
    import threading

    with dask.config.set(x=threading.Lock()):
        with dask.config.set(y=1):
            pass


@pytest.mark.parametrize("mkdir", [True, False])
def test_ensure_file_directory(mkdir, tmp_path):
    a = {"x": 1, "y": {"a": 1}}

    source = tmp_path / "source.yaml"
    dest = tmp_path / "dest"
    source.write_text(yaml.dump(a))

    if mkdir:
        dest.mkdir()

    ensure_file(source=source, destination=dest)

    assert dest.is_dir()
    assert (dest / "source.yaml").exists()


def test_ensure_file_defaults_to_DASK_CONFIG_directory(tmp_path):
    a = {"x": 1, "y": {"a": 1}}
    source = tmp_path / "source.yaml"
    source.write_text(yaml.dump(a))

    destination = tmp_path / "dask"
    PATH = dask.config.PATH
    try:
        dask.config.PATH = destination
        ensure_file(source=source)
    finally:
        dask.config.PATH = PATH

    assert destination.is_dir()
    [fn] = os.listdir(destination)
    assert os.path.split(fn)[1] == os.path.split(source)[1]


def test_pop():
    config = {"foo": {"ba-r": 1, "baz": 2}, "asd": 3}
    with pytest.raises(KeyError):
        pop("x", config=config)
    assert pop("x", default=4, config=config) == 4
    assert pop("foo.ba_r", config=config) == 1
    assert pop("asd", config=config) == 3
    assert config == {"foo": {"baz": 2}}


def test_refresh():
    defaults = []
    config = {}

    update_defaults({"a": 1}, config=config, defaults=defaults)
    assert config == {"a": 1}

    refresh(paths=[], env={"DASK_B": "2"}, config=config, defaults=defaults)
    assert config == {"a": 1, "b": 2}

    refresh(paths=[], env={"DASK_C": "3"}, config=config, defaults=defaults)
    assert config == {"a": 1, "c": 3}


@pytest.mark.parametrize(
    "inp,out",
    [
        ("1", "1"),
        (1, 1),
        ("$FOO", "foo"),
        ([1, "$FOO"], [1, "foo"]),
        ((1, "$FOO"), (1, "foo")),
        ({1, "$FOO"}, {1, "foo"}),
        ({"a": "$FOO"}, {"a": "foo"}),
        ({"a": "A", "b": [1, "2", "$FOO"]}, {"a": "A", "b": [1, "2", "foo"]}),
    ],
)
def test_expand_environment_variables(monkeypatch, inp, out):
    monkeypatch.setenv("FOO", "foo")
    assert expand_environment_variables(inp) == out


def test_env_var_canonical_name(monkeypatch):
    value = 3
    monkeypatch.setenv("DASK_A_B", str(value))
    d = {}
    dask.config.refresh(config=d)
    assert get("a_b", config=d) == value
    assert get("a-b", config=d) == value


def test_get_set_canonical_name():
    c = {"x-y": {"a_b": 123}}

    keys = ["x_y.a_b", "x-y.a-b", "x_y.a-b"]
    for k in keys:
        assert dask.config.get(k, config=c) == 123

    with dask.config.set({"x_y": {"a-b": 456}}, config=c):
        for k in keys:
            assert dask.config.get(k, config=c) == 456

    # No change to new keys in sub dicts
    with dask.config.set({"x_y": {"a-b": {"c_d": 1}, "e-f": 2}}, config=c):
        assert dask.config.get("x_y.a-b", config=c) == {"c_d": 1}
        assert dask.config.get("x_y.e_f", config=c) == 2


@pytest.mark.parametrize("key", ["custom_key", "custom-key"])
def test_get_set_roundtrip(key):
    value = 123
    with dask.config.set({key: value}):
        assert dask.config.get("custom_key") == value
        assert dask.config.get("custom-key") == value


def test_merge_None_to_dict():
    assert dask.config.merge({"a": None, "c": 0}, {"a": {"b": 1}}) == {
        "a": {"b": 1},
        "c": 0,
    }


def test_core_file():
    assert "temporary-directory" in dask.config.config
    assert "dataframe" in dask.config.config
    assert "compression" in dask.config.get("dataframe.shuffle")


def test_schema():
    jsonschema = pytest.importorskip("jsonschema")

    root_dir = pathlib.Path(__file__).parent.parent
    config = yaml.safe_load((root_dir / "dask.yaml").read_text())
    schema = yaml.safe_load((root_dir / "dask-schema.yaml").read_text())

    jsonschema.validate(config, schema)


def test_schema_is_complete():
    root_dir = pathlib.Path(__file__).parent.parent
    config = yaml.safe_load((root_dir / "dask.yaml").read_text())
    schema = yaml.safe_load((root_dir / "dask-schema.yaml").read_text())

    def test_matches(c, s):
        for k, v in c.items():
            if list(c) != list(s["properties"]):
                raise ValueError(
                    "\nThe dask.yaml and dask-schema.yaml files are not in sync.\n"
                    "This usually happens when we add a new configuration value,\n"
                    "but don't add the schema of that value to the dask-schema.yaml file\n"
                    "Please modify these files to include the missing values: \n\n"
                    "    dask.yaml:        {}\n"
                    "    dask-schema.yaml: {}\n\n"
                    "Examples in these files should be a good start, \n"
                    "even if you are not familiar with the jsonschema spec".format(
                        sorted(c), sorted(s["properties"])
                    )
                )
            if isinstance(v, dict):
                test_matches(c[k], s["properties"][k])

    test_matches(config, schema)


def test_rename():
    aliases = {
        "foo-bar": "foo.bar",
        "x.y": "foo.y",
        "a.b": "ab",
        "not-found": "not.found",
        "super.old": "super",
    }
    config = {
        "foo_bar": 1,
        "x": {"y": 2, "z": 3},
        "a": {"b": None},
        "super": {"old": 4},
    }
    expect = {"a.b": "ab", "foo-bar": "foo.bar", "super.old": "super", "x.y": "foo.y"}
    actual = rename(aliases, config=config)
    assert actual == expect
    assert config == {
        "foo": {"bar": 1, "y": 2},
        "x": {"z": 3},
        "a": {},
        "ab": None,
        "super": 4,
    }


@pytest.mark.parametrize(
    "args,kwargs",
    [
        ((), {"fuse_ave_width": 123}),
        (({"fuse_ave_width": 123},), {}),
        (({"fuse-ave-width": 123},), {}),
    ],
)
def test_deprecations_on_set(args, kwargs):
    with pytest.warns(FutureWarning, match=r"optimization\.fuse\.ave-width"):
        with dask.config.set(*args, **kwargs):
            assert dask.config.get("optimization.fuse.ave-width") == 123


def test_deprecations_on_env_variables(monkeypatch):
    d = {}
    monkeypatch.setenv("DASK_FUSE_AVE_WIDTH", "123")
    with pytest.warns(FutureWarning, match=r"optimization\.fuse\.ave-width"):
        dask.config.refresh(config=d)
    assert get("optimization.fuse.ave-width", config=d) == 123


@pytest.mark.parametrize("key", ["fuse-ave-width", "fuse_ave_width"])
def test_deprecations_on_yaml(tmp_path, key):
    d = {}
    (tmp_path / "dask.yaml").write_text(yaml.dump({key: 123}))

    with pytest.warns(FutureWarning, match=r"optimization\.fuse\.ave-width"):
        dask.config.refresh(config=d, paths=[tmp_path], env={})
    assert get("optimization.fuse.ave-width", config=d) == 123


def test_deprecations_warn_once(tmp_path):
    """When aggregating multiple YAML files, only warn once per deprecated key"""
    d = {}
    (tmp_path / "1.yaml").write_text(yaml.dump({"fuse-ave-width": 123}))
    (tmp_path / "2.yaml").write_text(yaml.dump({"fuse-ave-width": 123}))
    with pytest.warns(FutureWarning, match=r"optimization\.fuse\.ave-width") as info:
        dask.config.refresh(config=d, paths=[tmp_path], env={})
    assert len(info) == 1


def test_deprecations_suppressed_by_newer_yaml(tmp_path):
    """When a deprecated key is present in a config file, suppress the warning
    if the key is explicitly overridden in a higher-priority config file. This supports
    deployments where the root user installed an older version of Dask and the local
    user wants to use a newer one in a local environment.
    """
    d = {}
    # Note: within the same directory config files are always read in alphabetical order
    (tmp_path / "1.yaml").write_text(
        yaml.dump({"fuse-ave-width": 123, "fuse-max-width": 456})
    )
    (tmp_path / "2.yaml").write_text(
        # Suppress warning for fuse-ave-width in lower-priority config file
        yaml.dump({"optimization": {"fuse": {"ave-width": 789}}})
    )
    with pytest.warns(FutureWarning, match=r"optimization\.fuse\.max-width"):
        dask.config.refresh(config=d, defaults=[], paths=[tmp_path], env={})
    assert d == {
        "optimization": {
            "fuse": {
                "ave-width": 789,
                "max-width": 456,
            }
        }
    }


def test_deprecations_deleted_keys(tmp_path, monkeypatch):
    """When the user sets a deleted deprecated key with dask.config.key, raise
    ValueError. When the key is present in a config file, emit a warning and remove it
    from the config. However, if the key is explicitly overridden with a
    `ignore-deprecated` prefix in a higher-priority config, suppress the warning.
    """
    # At the moment of writing there are no deleted keys
    monkeypatch.setitem(dask.config.deprecations, "optimization.deleted-key", None)

    d = {}
    with pytest.raises(ValueError, match=r"optimization\.deleted-key"):
        dask.config.set({"optimization.deleted-key": 123}, config=d)
    assert not d

    (tmp_path / "1.yaml").write_text(
        yaml.dump(
            {"optimization": {"fuse": {"ave-width": 123}, "deleted-key": 456}},
        )
    )
    with pytest.warns(
        FutureWarning, match=r"ignore-deprecated\.optimization\.deleted-key"
    ):
        dask.config.refresh(config=d, defaults=[], paths=[tmp_path], env={})
    assert d == {"optimization": {"fuse": {"ave-width": 123}}}

    (tmp_path / "2.yaml").write_text(
        yaml.dump(
            {"ignore-deprecated": {"optimization": {"deleted-key": 456}}},
        )
    )
    d = {}
    dask.config.refresh(config=d, defaults=[], paths=[tmp_path], env={})
    assert d == {
        "ignore-deprecated": {"optimization": {"deleted-key": 456}},
        "optimization": {"fuse": {"ave-width": 123}},
    }


def test_get_override_with():
    with dask.config.set({"foo": "bar"}):
        # If override_with is None get the config key
        assert dask.config.get("foo") == "bar"
        assert dask.config.get("foo", override_with=None) == "bar"

        # Otherwise pass the default straight through
        assert dask.config.get("foo", override_with="baz") == "baz"
        assert dask.config.get("foo", override_with=False) is False
        assert dask.config.get("foo", override_with=True) is True
        assert dask.config.get("foo", override_with=123) == 123
        assert dask.config.get("foo", override_with={"hello": "world"}) == {
            "hello": "world"
        }
        assert dask.config.get("foo", override_with=["one"]) == ["one"]


def test_config_serialization():
    # Use context manager without changing the value to ensure test side effects are restored
    with dask.config.set({"array.svg.size": dask.config.get("array.svg.size")}):
        # Take a round trip through the serialization
        serialized = serialize({"array": {"svg": {"size": 150}}})
        config = deserialize(serialized)

        dask.config.update(dask.config.global_config, config)
        assert dask.config.get("array.svg.size") == 150


def test_config_inheritance():
    config = collect_env(
        {"DASK_INTERNAL_INHERIT_CONFIG": serialize({"array": {"svg": {"size": 150}}})}
    )
    assert dask.config.get("array.svg.size", config=config) == 150


def test__get_paths(monkeypatch):
    # These settings are used by Dask's config system. We temporarily
    # remove them to avoid interference from the machine where tests
    # are being run.
    monkeypatch.delenv("DASK_CONFIG", raising=False)
    monkeypatch.delenv("DASK_ROOT_CONFIG", raising=False)
    monkeypatch.setattr(site, "PREFIXES", [])

    expected = [
        "/etc/dask",
        os.path.join(sys.prefix, "etc", "dask"),
        os.path.join(os.path.expanduser("~"), ".config", "dask"),
    ]
    paths = _get_paths()
    assert paths == expected
    assert len(paths) == len(set(paths))  # No duplicate paths

    with monkeypatch.context() as m:
        m.setenv("DASK_CONFIG", "foo-bar")
        paths = _get_paths()
        assert paths == expected + ["foo-bar"]
        assert len(paths) == len(set(paths))

    with monkeypatch.context() as m:
        m.setenv("DASK_ROOT_CONFIG", "foo-bar")
        paths = _get_paths()
        assert paths == ["foo-bar"] + expected[1:]
        assert len(paths) == len(set(paths))

    with monkeypatch.context() as m:
        prefix = os.path.join("include", "this", "path")
        m.setattr(site, "PREFIXES", site.PREFIXES + [prefix])
        paths = _get_paths()
        assert os.path.join(prefix, "etc", "dask") in paths
        assert len(paths) == len(set(paths))


def test_default_search_paths():
    # Ensure _get_paths() is used for default paths
    assert dask.config.paths == _get_paths()


def test_config_files_order(tmp_path, monkeypatch):
    """Test that config files in different directories are read in priority order, while
    within the the same directory are read in alphabetical order, regardless of what
    order the syscalls return them in. os.listdir _typically_ returns results in
    alphabetical order but there is no guarantee in POSIX for it.
    """
    for i in range(9):
        (tmp_path / f"{i}.yaml").write_text(yaml.dump({"k": i}))
    (tmp_path / "a").mkdir()
    (tmp_path / "a" / "b.yaml").write_text(yaml.dump({"k": 10}))

    listdir_orig = os.listdir
    # Test that we're monkey-patching the right function
    monkeypatch.setattr(os, "listdir", lambda _: [])
    assert list(collect_yaml([tmp_path])) == []

    def shuffled_listdir(path):
        res = listdir_orig(path)
        random.shuffle(res)
        return res

    monkeypatch.setattr(os, "listdir", shuffled_listdir)

    actual = list(collect_yaml([tmp_path, tmp_path / "a"]))
    expect = [{"k": i} for i in range(9)] + [{"k": 10}]
    assert actual == expect

    actual = list(collect_yaml([tmp_path / "a", tmp_path]))
    expect = [{"k": 10}] + [{"k": i} for i in range(9)]
    assert actual == expect
